import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.image.MemoryImageSource;

/**
 * This class implements the map of the game.
 */
class Map {
	private int[] map;
	private int maxx = 0, maxy = 0;
	private int cellsFilled = 0;
	private int dWall;
	private boolean dirty = false;

	// buffer for floodfill
	private int[] Bufferx = null;
	private int[] Buffery = null;
	private int bufstart = 0;
	private int bufend = 0;

	/**
	 * Free cell in the map.
	 */
	public final static int FREE = Color.black.getRGB();
	/**
	 * Cell marked for flood_fill.
	 * 
	 * @see Map#flood_fill
	 */
	public final static int MARKED = Color.red.getRGB();
	/**
	 * Wall in the map.
	 */
	public final static int WALL = Color.green.getRGB();
	/**
	 * Path in the map.
	 */
	public final static int PATH = Color.yellow.getRGB();
	/**
	 * Cell inside captured space in the map.
	 */
	public final static int FILLED = Color.white.getRGB();

	/**
	 * Point is empty.
	 */
	public final static int EMPTY = 0;
	/**
	 * Point is outside the gamefield.
	 */
	public final static int OUTSIDE = 1;
	/**
	 * Point lies on the wall.
	 */
	public final static int COLLISION = 2;
	/**
	 * Point lies on the path.
	 */
	public final static int CROSSPATH = 3;

	/**
	 * Wall is vertical.
	 */
	public final static int WALL_VERTICAL = 1;
	/**
	 * Wall is horizontal.
	 */
	public final static int WALL_HORIZONTAL = 2;

	/**
	 * Map scaling coefficient.
	 */
	public final static int COEFF = 2;

	/**
	 * Create a map.
	 * 
	 * @param sizex
	 *            Map width.
	 * @param sizey
	 *            Map height.
	 * @param dw
	 *            Distance between external wall and the end of the map.
	 */
	public Map(int sizex, int sizey, int dw) {
		maxx = sizex / COEFF;
		maxy = sizey / COEFF;
		map = new int[maxx * maxy];
		dWall = dw / COEFF;
		clear();
		System.out.println("free is " + FREE);
		System.out.println("filled is " + FILLED);
		System.out.println("wall is " + WALL);
	}

	/**
	 * Clear the map. Only external walls remain.
	 */
	public void clear() {
		for (int i = 0; i < map.length; ++i)
			map[i] = FREE;
		// horizontal walls
		for (int i = dWall; i < maxx - dWall; ++i) {
			map[i + dWall * 3 * maxx] = WALL;
			map[i + maxx * (maxy - 1) - dWall * maxx] = WALL;
		}
		// vertical walls
		for (int i = dWall * 3; i < maxy - dWall; ++i) {
			map[maxx * i + dWall] = WALL;
			map[maxx * i + maxy - 1 - dWall] = WALL;
		}
		dirty = true;
	}

	private int _check(int x, int y) {
		return map[maxx * y + x];
	}

	/**
	 * Check what is in this point on the map.
	 */
	public int check(Point p) {
		return _check(p.x / COEFF, p.y / COEFF);
	}

	/**
	 * Check what is in this point on the map.
	 */
	public int check(int x, int y) {
		return _check(x / COEFF, y / COEFF);
	}

	private void _set(int x, int y, int val) {
		map[maxx * y + x] = val;
	}

	/**
	 * Set the value on the map.
	 */
	public void set(Point p, int val) {
		_set(p.x / COEFF, p.y / COEFF, val);
	}

	/**
	 * Set the value on the map.
	 */
	public void set(int x, int y, int val) {
		_set(x / COEFF, y / COEFF, val);
	}

	/**
	 * Get filled percentage.
	 */
	public int percent() {
		return cellsFilled * 100 / ((maxx - 2 * dWall) * (maxy - 4 * dWall));
	}

	/**
	 * Get filled space area.
	 */
	public int filled() {
		return cellsFilled;
	}

	/**
	 * Clean current path from the map.
	 */
	public void wipe_path() {
		for (int x = dWall; x < maxx - dWall; ++x)
			for (int y = dWall * 3; y < maxy - dWall; ++y)
				if (_check(x, y) == PATH) {
					_set(x, y, FREE);
				}
		dirty = true;
	}

	/**
	 * Fill all the free space with the value <code>FILLED</code>.
	 */
	public void fill_all() {
		for (int x = dWall; x < maxx - dWall; ++x)
			for (int y = dWall * 3; y < maxy - dWall; ++y) {
				if (_check(x, y) == FREE) {
					cellsFilled++;
					_set(x, y, FILLED);
				}
				if (_check(x, y) == PATH) {
					_set(x, y, WALL);
				}
			}

		dirty = true;
	}

	private void flood_fill_cell(int x, int y) {
		map[maxx * y + x] = FREE;
		cellsFilled--;
	}

	private void put_on_stack(int x, int y) {
		if (x < dWall || y < dWall * 3 || x >= maxx - dWall
				|| y >= maxy - dWall)
			return;
		if (map[maxx * y + x] != FILLED)
			return;
		map[maxx * y + x] = MARKED;
		Bufferx[bufend] = x;
		Buffery[bufend++] = y;
	}

	/**
	 * Frees all the space in the area containing the point.
	 */
	public void flood_fill(Point p) {
		Bufferx = new int[maxx * maxy];
		Buffery = new int[maxx * maxy];
		bufstart = bufend = 0;
		put_on_stack(p.x / 2, p.y / 2); // !!!
		int x, y;
		while (bufstart != bufend) {
			x = Bufferx[bufstart];
			y = Buffery[bufstart++];
			if (map[maxx * y + x] != MARKED)
				continue;
			flood_fill_cell(x, y);
			// upper point
			put_on_stack(x, y - 1);
			// lower point
			put_on_stack(x, y + 1);
			// left point
			put_on_stack(x - 1, y);
			// right point
			put_on_stack(x + 1, y);
		}
		Bufferx = Buffery = null;
		System.gc();
		dirty = true;
	}

	/**
	 * Checks what means the map value in the point
	 * 
	 * @see Map#OUTSIDE
	 * @see Map#EMPTY
	 * @see Map#CROSSPATH
	 * @see Map#COLLISION
	 */
	public int position(Point p) {
		int x = p.x / COEFF;
		int y = p.y / COEFF;
		if (x < dWall || x >= maxx - dWall || y < dWall * 3
				|| y >= maxx - dWall)
			return OUTSIDE;
		if (map[x + y * maxx] == FILLED)
			return OUTSIDE;
		if (map[x + y * maxx] == FREE)
			return EMPTY;
		if (map[x + y * maxx] == PATH)
			return CROSSPATH;
		if (map[x + y * maxx] == WALL)
			return COLLISION;
		return EMPTY;
	}

	/**
	 * Finds if the Movable object has certain value in the map under it.
	 * 
	 * @return <strong>true</strong> if the value found,<br>
	 *         <strong>false</strong> otherwise. <code>collp</code> is set to
	 *         the first point where the value is added.
	 */
	public boolean find(Movable m, Point collp, int what) {
		Point c = m.get_center();
		Point tryp = new Point(0, 0);
		for (int i = c.x - Movable.radius; i < c.x + Movable.radius; ++i)
			for (int j = c.y - Movable.radius; j < c.y + Movable.radius; ++j) {
				tryp.move(i, j);
				if (position(tryp) == what) {
					collp.move(i, j);
					return true;
				}
			}
		return false;
	}

	/**
	 * Rounds the value to be multiplicate of the scale coeeficient.
	 */
	public static int round(int rounded) {
		return rounded / COEFF * COEFF;
	}

	/**
	 * Guesses wall direction in the point.
	 * 
	 * @return Combination of the <code>WALL_VERTICAL</code> and
	 *         <code>WALL_HORIZONTAL</code> flags.
	 * @see WALL_VERTICAL
	 * @see WALL_HORIZONTAL
	 */
	public int wallDirection(Point p) {
		Point tryp = new Point(p.x, p.y);
		int result = 0;
		if (position(tryp) != COLLISION)
			return result;

		tryp.move(p.x - COEFF, p.y);
		if (position(tryp) == COLLISION)
			result |= WALL_HORIZONTAL;

		tryp.translate(p.x + COEFF, p.y);
		if (position(tryp) == COLLISION)
			result |= WALL_HORIZONTAL;

		tryp.move(p.x, p.y - COEFF);
		if (position(tryp) == COLLISION)
			result |= WALL_VERTICAL;

		tryp.translate(p.x, p.y + COEFF);
		if (position(tryp) == COLLISION)
			result |= WALL_VERTICAL;

		return result;
	}

	/**
	 * Updates rectangle on the map.
	 */
	public void updateRect(Graphics g, int startx, int starty, int endx,
			int endy, Xonix parent) {
		if (startx < 0)
			startx = 0;
		if (starty < 0)
			starty = 0;
		if (endx >= maxx)
			endx = maxx * COEFF;
		if (endy >= maxy)
			endy = maxy * COEFF;
		System.out.println(startx + " " + starty + " " + endx + " " + endy);
		Image img = parent.createImage(new MemoryImageSource((endx - startx)
				/ COEFF, (endy - starty) / COEFF, map, startx / COEFF + starty
				/ COEFF * maxx, maxx));
		// parent.imageUpdate(img,0,startx,starty,endx-startx+1,endy-starty+1);
		// System.out.println("wait");
		// while(!parent.prepareImage(img,parent));
		// System.out.println("done");
		g.drawImage(img, startx, starty, endx - startx, endy - starty, parent);
		// System.out.println("done draw");
	}

	/**
	 * Draws the map.
	 */
	public void paint(Graphics g, Xonix parent) {
		dirty = true;
		update(g, parent);
	}

	/**
	 * Updates the map.
	 */
	public void update(Graphics g, Xonix parent) {
		if (!dirty)
			return;
		updateRect(g, 0, 0, maxx * COEFF, maxy * COEFF, parent);
		dirty = false;
	}
}
